@@ -2,10 +2,12 @@
 #' git protocol notes, for developers
 #'
 #' Assumptions, they might be relaxed or checked for later:
-#' - The server must speak the smart protocol, version 2.
+#' - The server must speak the smart protocol, version 1 or 2.
+#'   (Although I added some functions specifically for the dumb protocol,
+#'   but they are pretty limited.)
 #' - We use HTTP transport, not SSH.
 #' - The server should have the `shallow` capability.
-#' - The server should have the `filter` capability.
+#' - The server should have the `filter` capability if protocol version 2.
 #' - Only SHA-1 hashing is supported.
 #'
 #' Improvements needed:
@@ -84,11 +86,7 @@ git_list_refs <- function(url, prefixes = NULL) {
 }
 
 async_git_list_refs <- function(url, prefixes = NULL) {
-  if (is.null(prefixes)) {
-    async_git_list_refs_v1(url)
-  } else {
-    async_git_list_refs_v2(url, prefixes)
-  }
+  async_git_list_refs_v2(url, prefixes)
 }
 
 # -------------------------------------------------------------------------
@@ -136,7 +134,7 @@ async_git_resolve_ref <- function(url, ref) {
     }
     async_git_list_refs(url, filt)$
       then(function(refs) {
-        if (ref %in% refs$refs$ref) {
+        result <- if (ref %in% refs$refs$ref) {
           refs$refs$hash[refs$refs$ref == ref]
 
         } else if (paste0("refs/tags/", ref) %in% refs$refs$ref) {
@@ -146,7 +144,7 @@ async_git_resolve_ref <- function(url, ref) {
           refs$refs$hash[refs$refs$ref == paste0("refs/heads/", ref)]
 
         } else if (any(startsWith(refs$refs$hash, ref))) {
-          sha <- refs$refs$hash[startsWith(refs$refs$hash, ref)]
+          sha <- unique(refs$refs$hash[startsWith(refs$refs$hash, ref)])
           if (length(sha) > 1) {
             throw(pkg_error(
               "Found multiple git refs with prefix {.val {ref}}, it is ambiguous.",
@@ -159,13 +157,13 @@ async_git_resolve_ref <- function(url, ref) {
         } else {
           throw(pkg_error(
             "Unknown git ref: {.val {ref}}.",
-            "i" = if (grepl("^[0-9a-f]+$", ref)) {
-              "If you want to specify a SHA prefix, then use at least 7 hexa digits."
-             },
             .class = "git_proto_error_unknown_ref",
             .data = list(ref = ref, url = redact_url(url))
           ))
         }
+
+        attr(result, "protocol") <- if ("version 2" %in% refs$caps) 2 else 1
+        result
       })
 
   } else {
@@ -187,8 +185,7 @@ async_git_list_files_process <- function(packfile, ref, sha) {
   tree_sizes <- viapply(packfile, function(x) nrow(x$object) %||% NA_integer_)
   num_files <- sum(tree_sizes, na.rm = TRUE)
 
-  files = data.frame(
-    stringsAsFactors = FALSE,
+  files = data_frame(
     hash = character(num_files),
     type = character(num_files),
     mode = character(num_files),
@@ -323,7 +320,34 @@ git_fetch <- function(url, sha, blobs = FALSE) {
 async_git_fetch <- function(url, sha, blobs = FALSE) {
   url; sha
 
-  async_git_send_message(
+  if (!is.null(attr(sha, "protocol"))) {
+    if (attr(sha, "protocol") == 1) {
+      async_git_fetch_v1(url, sha, blobs)
+    } else {
+      async_git_fetch_v2(url, sha, blobs)
+    }
+  } else {
+    async_git_fetch_v2(url, sha, blobs)$
+      catch(error = function(err) async_git_fetch_v1(url, sha, blobs))
+  }
+}
+
+async_git_fetch_v1 <- function(url, sha, blobs) {
+  async_git_send_message_v1(
+    url,
+    args = c(
+      paste0("want ", sha),
+      "deepen 1",
+      "",
+      "done",
+      ""
+    ),
+    caps = c("no-done", "no-progress", paste0("agent=", git_ua()))
+  )$then(function(reply) git_fetch_process_v1(reply, url, sha))
+}
+
+async_git_fetch_v2 <- function(url, sha, blobs) {
+  async_git_send_message_v2(
     url,
     "fetch",
     caps = c(
@@ -341,6 +365,21 @@ async_git_fetch <- function(url, sha, blobs = FALSE) {
   )$then(function(reply) git_fetch_process(reply, url, sha))
 }
 
+git_fetch_process_v1 <- function(reply, url, sha) {
+  if (length(reply) == 0) {
+    throw(pkg_error(
+      "Empty reply from git server (protocol v1) at {.url {redact_url(url)}}."
+    ))
+  }
+  if (reply[[length(reply)]]$type != "pack") {
+    throw(pkg_error(
+      "No PACK in git server response (protocol v1) from {.url {redact_url(url)}}."
+    ))
+  }
+
+  git_unpack(reply[[length(reply)]]$data)
+}
+
 git_fetch_process <- function(reply, url, sha) {
 
   # https://github.com/git/git/blob/7c2ef319c52c4997256f5807564523dfd4acdfc7/Documentation/gitprotocol-v2.txt#L240
@@ -484,10 +523,10 @@ unpack_packfile_repo <- function(parsed, output) {
       if (tr$type[l] == "tree") {
         tidx <- which(tr$hash[l] == names(trees))[1]
         if (is.na(tidx)) {
-          throw(new_error(
-            "git tree missing from packfile: {.val {tr$hash[i]}}.",
-            .class = "git_proto_error_unexpected_response"
-          ))
+          throw(new_error(                                          # nocov
+            "git tree missing from packfile: {.val {tr$hash[i]}}.", # nocov
+            .class = "git_proto_error_unexpected_response"          # nocov
+          ))                                                        # nocov
         }
         wd <<- c(wd, tr$path[l])
         mkdirp(opath)
@@ -563,15 +602,23 @@ git_parse_message <- function(msg) {
       ))
     }
     pkg_len <- msg[pos:(pos+3)]
-    if (any(pkg_len < charToRaw('0') | pkg_len > charToRaw('f'))) {
-      throw(pkg_error(
-        "Invalid pkt-len field in message from git, must be four hexa digits.",
-        .class = "git_proto_error_invalid_data"
-      ))
+    if (!all(pkg_len == charToRaw("PACK"))) {
+      if (any(pkg_len < charToRaw('0') | pkg_len > charToRaw('f'))) {
+        throw(pkg_error(
+          "Invalid pkt-len field in message from git, must be four hexa digits.",
+          .class = "git_proto_error_invalid_data"
+        ))
+      }
+      len <- as.integer(as.hexmode(rawToChar(pkg_len)))
     }
-    len <- as.integer(as.hexmode(rawToChar(pkg_len)))
 
-    if (len == 0) {
+    if (all(pkg_len == charToRaw("PACK"))) {
+      lines[[length(lines) + 1L]] <- list(
+        type = "pack",
+        data = msg[pos:length(msg)]
+      )
+      pos <- length(msg) + 1L
+    } else if (len == 0) {
       lines[[length(lines) + 1L]] <- list(type = "flush-pkt")
       pos <- pos + 4L
     } else if (len == 1) {
@@ -653,6 +700,25 @@ git_create_message_v2 <- function(
   )
 }
 
+git_create_message_v1 <- function(args = character(), caps = character()) {
+  if (length(args) == 0) {
+    throw(pkg_error(
+      "Invalid git protocol (v1) message, must have at least one argument"
+    ))
+  }
+
+  args[1] <- paste0(args[1], " ", paste(caps, collapse = " "))
+  args <- ifelse(
+    last_char(args) == "\n" | args == "",
+    args,
+    paste0(args, "\n")
+  )
+
+  res <- lapply(args, pkt_line)
+  res[args == ""] <- list(charToRaw("0000"))
+  unlist(res)
+}
+
 #' Send a protocol version 2 message to a git server
 #'
 #' @inheritParams git_list_refs
@@ -661,21 +727,21 @@ git_create_message_v2 <- function(
 #'
 #' @noRd
 
-git_send_message <- function(
+git_send_message_v2 <- function(
   url,
   cmd,
   caps = character(),
   args = character()) {
 
-  synchronize(async_git_send_message(url, cmd, caps = caps, args = args))
+  synchronize(async_git_send_message_v2(url, cmd, caps = caps, args = args)) # nocov
 }
 
-#' `async_git_send_message()` is the asynchronous variant of
-#' `git_send_message()`.
-#' @rdname git_send_message
+#' `async_git_send_message_v2()` is the asynchronous variant of
+#' `git_send_message_v2()`.
+#' @rdname git_send_message_v2
 #' @noRd
 
-async_git_send_message <- function(
+async_git_send_message_v2 <- function(
   url,
   cmd,
   caps = character(),
@@ -700,6 +766,25 @@ async_git_send_message <- function(
     then(function(res) git_parse_message(res$content))
 }
 
+async_git_send_message_v1 <- function(url, args, caps) {
+
+  msg <- git_create_message_v1(caps = caps, args = args)
+
+  url2 <- paste0(url, "/git-upload-pack")
+  headers <- c(
+    "Content-Type" = "application/x-git-upload-pack-request",
+    "User-Agent" = git_ua(),
+    "accept" = "application/x-git-upload-pack-result",
+    "content-length" = as.character(length(msg))
+  )
+  http_post(
+    url2,
+    data = msg,
+    headers = headers
+  )$then(http_stop_for_status)$
+    then(function(res) git_parse_message(res$content))
+}
+
 delim_pkt <- function() {
   charToRaw("0001")
 }
@@ -779,41 +864,59 @@ async_git_list_refs_v1 <- function(url) {
 
 git_list_refs_v1_process <- function(response, url) {
   psd <- git_parse_message(response$content)
+  psd <- drop_service_line(psd)
   check_initial_response(psd, url)
 
   # The stream MUST include capability declarations behind a NUL on the
   # first ref.
-  nul <- which(psd[[3]]$data == 0)
+  nul <- which(psd[[1]]$data == 0)
   if (length(nul) != 1) {
-      throw(pkg_error(
-        "Unexpected response from git server, third pkt-line should contain
-         exactly one {.code NUL} byte.",
-        .class = "git_proto_error_unexpected_response",
-        .data = list(url = redact_url(url))
-      ))
+    # nocov start
+    throw(pkg_error(
+      "Unexpected response from git server, third pkt-line should contain
+       exactly one {.code NUL} byte.",
+      .class = "git_proto_error_unexpected_response",
+      .data = list(url = redact_url(url))
+    ))
+    # nocov end
   }
 
-  caps <- if (nul < length(psd[[3]]$data)) {
-    rawToChar(psd[[3]]$data[(nul+1):length(psd[[3]]$data)])
+  caps <- if (nul < length(psd[[1]]$data)) {
+    rawToChar(psd[[1]]$data[(nul+1):length(psd[[1]]$data)])
   } else {
     ""                                                              # nocov
   }
   caps <- strsplit(trimws(caps), " ", fixed = TRUE)[[1]]
-  psd[[3]]$data <- if (nul == 1) raw() else psd[[3]]$data[1:(nul-1)]
-  psd[[3]]$text <- rawToChar(psd[[3]]$data)
+  psd[[1]]$data <- if (nul == 1) raw() else psd[[1]]$data[1:(nul-1)]
+  psd[[1]]$text <- rawToChar(psd[[1]]$data)
 
   if (psd[[length(psd)]]$type != "flush-pkt") {
+    # nocov start
     throw(pkg_error(
       "Response from git server does not have a closing {.code flush-pkt}.",
       .class = "git_proto_error_unexpected_response",
       .data = list(url = redact_url(url))
     ))
+    # nocov end
   }
 
-  refs <- git_parse_pkt_line_refs(psd[3:(length(psd)-1)], url)
+  refs <- git_parse_pkt_line_refs(psd[1:(length(psd)-1)], url)
   list(refs = refs, caps = caps)
 }
 
+git_list_refs_v1_process_1 <- function(response, url, prefixes) {
+  # We wanted v2, but the server replied in v1, so we'll use v1 now
+  refs <- git_list_refs_v1_process(response, url)
+  if (length(prefixes) > 0) {
+    keep <- logical(length(refs$refs$ref))
+    for (pfx in prefixes) {
+      keep <- keep | startsWith(refs$refs$ref, pfx)
+    }
+    refs$refs <- refs$refs[keep, ]
+  }
+  refs
+}
+
 #' Helper function to parse a `pkt_line` containing a git ref
 #'
 #' E.g. an `ls-refs` commans responds with such lines.
@@ -825,8 +928,7 @@ git_list_refs_v1_process <- function(response, url) {
 #' @noRd
 
 git_parse_pkt_line_refs <- function(lines, url) {
-  res <- data.frame(
-    stringsAsFactors = FALSE,
+  res <- data_frame(
     ref = character(length(lines)),
     hash = character(length(lines))
   )
@@ -835,12 +937,14 @@ git_parse_pkt_line_refs <- function(lines, url) {
     line <- lines[[idx]]
     if (line$type != "data-pkt" || is.null(line$text) ||
         !grepl("^[0-9a-f]{40} .+$", line$text)) {
+      # nocov start
       throw(pkg_error(
         "Expected response from git server.",
         i = "Line {idx} must be a text {.code data-pkt} with a sha and a ref name.",
         .class = "git_proto_error_unexpected_response",
         .data = list(url = redact_url(url))
       ))
+      # nocov end
     }
 
     res$ref[idx] <- substr(line$text, 42, nchar(line$text))
@@ -896,8 +1000,13 @@ async_git_list_refs_v2 <- function(url, prefixes = character()) {
 
 async_git_list_refs_v2_process_1 <- function(response, url, prefixes) {
   psd <- git_parse_message(response$content)
+  psd <- drop_service_line(psd)
   check_initial_response(psd, url)
+  async_git_list_refs_v2_process_2(response, psd, url, prefixes)
+}
 
+
+async_git_list_refs_v2_process_2 <- function(response, psd, url, prefixes) {
   # capability-advertisement = protocol-version
   #                capability-list
   #                flush-pkt
@@ -909,21 +1018,19 @@ async_git_list_refs_v2_process_1 <- function(response, url, prefixes) {
   # key = 1*(ALPHA | DIGIT | "-_")
   # value = 1*(ALPHA | DIGIT | " -_.,?\/{}[]()<>!@#$%^&*+=:;")
 
-  if (is.null(psd[[3]]$text)) {
+  if (is.null(psd[[1]]$text)) {
     throw(pkg_error(
-      "Invalid git protocol message from {.url {url}}."
+      "Invalid git protocol message from {.url {redact_url(url)}}."
     ))
   }
-  if (is.na(psd[[3]]$text)) {
-    throw(pkg_error(
-      "Only git protocol version 2 is supported.",
-      "i" = "{.url {url}} seems to support version 1 only.",
-      .class = "git_proto_error_not_implemented"
-    ))
+
+  if (is.na(psd[[1]]$text)) {
+    return(git_list_refs_v1_process_1(response, url, prefixes))
   }
-  if (psd[[3]]$text != "version 2") {
+
+  if (psd[[1]]$text != "version 2") {
     throw(pkg_error(
-      "Only git protocol version 2 is supported, not {psd[[3]]$text}.",
+      "Only git protocol version 2 is supported, not {psd[[1]]$text}.",
       .class = "git_proto_error_not_implemented"
     ))
   }
@@ -936,51 +1043,32 @@ async_git_list_refs_v2_process_1 <- function(response, url, prefixes) {
     ))
   }
 
-  caps <- unlist(lapply(psd[4:(length(psd)-1)], "[[", "text"))
+  caps <- unlist(lapply(psd[1:(length(psd)-1)], "[[", "text"))
 
   args <- if (length(prefixes)) paste0("ref-prefix ", prefixes, "\n")
 
-  async_git_send_message(url, "ls-refs", args = as.character(args))$
-    then(function(res) async_git_list_refs_v2_process_2(res, caps, url))
+  async_git_send_message_v2(url, "ls-refs", args = as.character(args))$
+    then(function(res) async_git_list_refs_v2_process_3(res, caps, url))
+}
+
+drop_service_line <- function(psd) {
+  # Actually, this might not be present in V2 servers...
+  if (length(psd) > 0 && !is.null(psd[[1]]$text) &&
+      grepl("^# service=.+", psd[[1]]$text)) {
+    psd <- psd[-1]
+  }
+
+  if (length(psd) > 0 && psd[[1]]$type == "flush-pkt") {
+    psd <- psd[-1]
+  }
+
+  psd
 }
 
 check_initial_response <- function(psd, url) {
-  # Clients MUST validate the first five bytes of the response entity
-  # matches the regex `^[0-9a-f]{4}#`.  If this test fails, clients
-  # MUST NOT continue.
-  if (length(psd) == 0 || is.null(psd[[1]]$text) ||
-      !grepl("^#", psd[[1]]$text)) {
+  if (length(psd) < 1 || psd[[1]]$type != "data-pkt") {
       throw(pkg_error(
-        "Unexpected response from git server, first pkt_line should start
-         with {.code #}.",
-        .class = "git_proto_error_unexpected_response",
-        .data = list(url = redact_url(url))
-      ))
-  }
-
-  # Clients MUST verify the first pkt-line is `# service=$servicename`.
-  if (!grepl("^# service=.+", psd[[1]]$text)) {
-      throw(pkg_error(
-        "Unexpected response from git server, first pkt_line should be
-         {.code #servicename}.",
-        .class = "git_proto_error_unexpected_response",
-        .data = list(url = redact_url(url))
-      ))
-  }
-
-  if (length(psd) < 2 || psd[[2]]$type != "flush-pkt") {
-      throw(pkg_error(
-        "Unexpected response from git server, second pkt_line should be a
-         {.code flush-pkt}.",
-        .class = "git_proto_error_unexpected_response",
-        .data = list(url = redact_url(url))
-      ))
-  }
-
-  if (length(psd) < 3 || psd[[3]]$type != "data-pkt") {
-      throw(pkg_error(
-        "Unexpected response from git server, third pkt-line should be a
-         {.code data-pkt}.",
+        "Unexpected response from git server, no {.code data-pkt} line.",
         .class = "git_proto_error_unexpected_response",
         .data = list(url = redact_url(url))
       ))
@@ -995,7 +1083,7 @@ check_initial_response <- function(psd, url) {
 #' @return Same as [git_list_refs()].
 #' @noRd
 
-async_git_list_refs_v2_process_2 <- function(reply, caps, url) {
+async_git_list_refs_v2_process_3 <- function(reply, caps, url) {
   if (reply[[length(reply)]]$type != "flush-pkt") {
     throw(pkg_error(
       "Response from git server does not have a closing {.code flush-pkt}.",
@@ -1070,26 +1158,36 @@ git_unpack <- function(pack) {
 
   idx <- 13L
   objects <- vector("list", n_obj)
+  object_starts <- integer()
 
   types <- c("commit", "tree", "blob", "tag", "reserved", "ofs_delta", "ref_delta")
 
   unpack_object <- function() {
+    start <- idx
     type <- bitwShiftR(bitwAnd(as.integer(pack[idx]), 0x7f), 4L)
     size <- parse_size(pack, idx)
     idx <<- size$idx + 1
-    if (type == 7L) {
+    if (type == 6L) {
+      offset <- parse_ofs_delta_offset(pack, idx)
+      idx <<- offset$idx + 1
+    } else if (type == 7L) {
       if (idx + 19 > length(pack)) {
+        # nocov start
         throw(pkg_error(
           "Invalid packfile, unexpected end of file"
         ))
+        # nocov end
       }
       base <- bin_to_sha(pack[idx:(idx+19)])
       idx <<- idx + 20
     }
     obj <- zip::inflate(pack, idx, size$size)
     idx <<- idx + obj$bytes_read
-    if (type == 7L) {
-      deltified_object(obj$output, base)
+    if (type == 6L) {
+      baseidx <- object_starts[[as.character(start - offset$size)]]
+      deltified_object(obj$output, baseidx = baseidx)
+    } else if (type == 7L) {
+      deltified_object(obj$output, base = base)
     } else {
       list(
         type = types[type],
@@ -1100,11 +1198,14 @@ git_unpack <- function(pack) {
     }
   }
 
-  deltified_object <- function(delta, base) {
-    baseidx <- match(base, names(objects))
-    if (is.na(baseidx)) {
-      throw(pkg_error(
-        "Invalid git packfile, cannot find base of ref-delta"
+  deltified_object <- function(delta, base = NULL, baseidx = NULL) {
+    baseidx <- baseidx %||% match(base, names(objects))
+    if (is.na(baseidx) || objects[[baseidx]]$type == "delta") {
+      return(list(
+        type = "delta",
+        data = delta,
+        base = base,
+        baseidx = if (is.na(baseidx)) NULL else baseidx
       ))
     }
     baseobj <- objects[[baseidx]]$raw
@@ -1139,33 +1240,137 @@ git_unpack <- function(pack) {
     )
   }
 
-  for (i in seq_len(n_obj)) {
-    objects[[i]] <- unpack_object()
-    if (objects[[i]]$type == "commit") {
-      objects[[i]]$object <- rawToChar(objects[[i]]$raw)
-    } else if (objects[[i]]$type == "tree") {
-      objects[[i]]$object <- parse_tree(objects[[i]]$raw)
+  finalize_object <- function(x) {
+    if (x$type == "commit") {
+      x$object <- rawToChar(x$raw)
+    } else if (x$type == "tree") {
+      x$object <- parse_tree(x$raw)
     }
-    if (objects[[i]]$type %in% c("commit", "tree", "blob", "tag")) {
+
+    if (x$type %in% c("commit", "tree", "blob", "tag")) {
       raw2 <- c(
-        charToRaw(paste0(objects[[i]]$type, " ", length(objects[[i]]$raw))),
+        charToRaw(paste0(x$type, " ", length(x$raw))),
         as.raw(0L),
-        objects[[i]]$raw
+        x$raw
       )
-      objects[[i]]$hash <- cli::hash_raw_sha1(raw2)
-      names(objects)[i] <- objects[[i]]$hash
+      x$hash <- cli::hash_raw_sha1(raw2)
+    } else if (x$type == "delta") {
+      # do nothing
     } else {
+      # nocov start
       throw(pkg_error(
-        "git packfile object type {.cls {objects[[i]]$type}} is not
+        "git packfile object type {.cls {x$type}} is not
          implemented yet.",
         .class = "git_proto_error_not_implemented"
       ))
+      # nocov end
     }
+    x
+  }
+
+  for (i in seq_len(n_obj)) {
+    object_starts[[as.character(idx)]] <- i
+    objects[[i]] <- unpack_object()
+    objects[[i]] <- finalize_object(objects[[i]])
+    if (!is.null(objects[[i]]$hash)) names(objects)[i] <- objects[[i]]$hash
+  }
+
+  # now need to resolve the deltas
+  n_delta <- sum(sapply(objects, "[[", "type") == "delta")
+  while (n_delta > 0) {
+    for (i in seq_len(n_obj)) {
+      if (objects[[i]]$type == "delta") {
+        objects[[i]] <- deltified_object(
+          objects[[i]]$data,
+          objects[[i]]$base,
+          objects[[i]]$baseidx
+        )
+        objects[[i]] <- finalize_object(objects[[i]])
+        if (!is.null(objects[[i]]$hash)) names(objects)[i] <- objects[[i]]$hash
+      }
+    }
+    n_delta2 <- sum(sapply(objects, "[[", "type") == "delta")
+    if (n_delta2 == n_delta) {
+      throw(pkg_error(
+        "Found circular references while resolving deltas in git pack file."
+      ))
+    }
+    n_delta <- n_delta2
   }
 
   objects
 }
 
+git_list_pack_index <- function(idx) {
+  if (is.character(idx)) {
+    idx <- readBin(idx, "raw", file.size(idx))
+  }
+
+  if (length(idx) < 4 + 4 + 256) {
+    throw(pkg_error(
+      "Invalid pack index file, too short: {length(idx)} byte{?s}."
+    ))
+  }
+
+  if (any(idx[1:4] != c(0xff, 0x74, 0x4f, 0x63))) {
+    throw(pkg_error(
+      "Invalid pack index file, no {.code \\377tOc} header."
+    ))
+  }
+  if (any(idx[5:8] != c(0, 0, 0, 2))) {
+    throw(pkg_error(
+      "Only version 2 pack index files are supported"
+    ))
+  }
+
+  tab <- matrix(as.integer(idx[9L:(9L + (256L * 4L) - 1L)]), nrow = 4)
+  tab <- tab[1,] * 256**3 + tab[2,] * 256**2 + tab[3,] * 256 + tab[4,]
+  n_obj <- tab[256]
+
+  hash_off <- 256L * 4L + 9L
+  hash_len <- n_obj * 20L
+  hash <- matrix(
+    as.character(idx[hash_off:(hash_off + hash_len - 1L)]),
+    nrow = 20
+  )
+  hash <- apply(hash, 2, paste, collapse = "")
+
+  crc_off <- hash_off + hash_len
+  crc_len <- n_obj * 4
+  crc <- matrix(as.integer(idx[crc_off:(crc_off + crc_len - 1L)]), nrow = 4)
+  crc <- crc[1,] * 256**3 + crc[2,] * 256**2 + crc[3,] * 256 + crc[4,]
+
+  off_off <- crc_off + crc_len
+  off_len <- n_obj * 4
+  off <- matrix(as.integer(idx[off_off:(off_off + off_len - 1L)]), nrow = 4)
+  off <- off[1,] * 256**3 + off[2,] * 256**2 + off[3,] * 256 + off[4,]
+
+  data_chksum <- idx[(off_off + off_len): (off_off + off_len + 20L - 1L)]
+  data_chksum <- paste(as.character(data_chksum), collapse = "")
+
+  idx_chksum <- idx[(off_off + off_len): (off_off + off_len + 20L - 1L) + 20L]
+  idx_chksum <- paste(as.character(idx_chksum), collapse = "")
+
+  if (length(idx) > off_off + off_len + 40L) {
+    warning("Ignored 8 byte offsets in git pack file")
+  }
+
+  objects <- data.frame(
+    stringsAsFactors = FALSE,
+    hash = hash,
+    crc = crc,
+    offset = off
+  )
+  objects <- objects[order(objects$offset), ]
+  rownames(objects) <- NULL
+
+  list(
+    objects = objects,
+    data_chksum = data_chksum,
+    idx_chksum = idx_chksum
+  )
+}
+
 #' Parse a four byte integer in network byte order
 #'
 #' @param x Raw vector or four bytes.
@@ -1210,10 +1415,12 @@ parse_size <- function(x, idx) {
   while (c >= 128) {
     idx <- idx + 1L
     if (idx > length(x)) {
+      # nocov start
       throw(pkg_error(
         "Invalid git packfile, invalid size field.",
         .class = "git_proto_error_invalid_data"
       ))
+      # nocov end
     }
     c <- as.integer(x[idx])
     size <- size + bitwShiftL(bitwAnd(c, 0x7f), shft)
@@ -1223,6 +1430,10 @@ parse_size <- function(x, idx) {
   list(size = size, idx = idx)
 }
 
+parse_ofs_size <- function(x, idx) {
+
+}
+
 parse_delta_size <- function(x, idx) {
   c <- as.integer(x[idx])
   size <- bitwAnd(c, 0x7f)
@@ -1230,10 +1441,12 @@ parse_delta_size <- function(x, idx) {
   while (c >= 128) {
     idx <- idx + 1L
     if (idx > length(x)) {
+      # nocov start
       throw(pkg_error(
         "Invalid git packfile, invalid size field.",
         .class = "git_proto_error_invalid_data"
       ))
+      # nocov end
     }
     c <- as.integer(x[idx])
     size <- size + bitwShiftL(bitwAnd(c, 0x7f), shft)
@@ -1243,6 +1456,19 @@ parse_delta_size <- function(x, idx) {
   list(size = size, idx = idx)
 }
 
+parse_ofs_delta_offset <- function(x, idx) {
+  c <- as.integer(x[idx])
+  size <- bitwAnd(c, 0x7f)
+  while (c >= 128) {
+    idx <- idx + 1L
+    c <- as.integer(x[idx])
+    size <- size + 1L
+    size <- bitwShiftL(size, 7) + (bitwAnd(c, 0x7f))
+  }
+
+  list(size = size, idx = idx)
+}
+
 parse_delta_offset <- function(x, idx) {
   c <- as.integer(x[idx])
   offset <- 0L
@@ -1309,8 +1535,7 @@ parse_tree <- function(tree) {
 
   nul <- nul[!is.na(nul)]
   num <- length(nul)
-  res <- data.frame(
-    stringsAsFactors = FALSE,
+  res <- data_frame(
     type = character(num),
     mode = character(num),
     path = character(num),
@@ -1354,10 +1579,12 @@ parse_commit <- function(commit) {
   lines <- strsplit(commit, "\n", fixed = TRUE)[[1]]
   delim <- which(lines == "")[1]
   if (delim == 1) {
+    # nocov start
     throw(pkg_error(
       "Invalid git commit object, no {.code tree} field.",
       .class = "git_proto_error_invalid_data"
     ))
+    # nocov end
   }
 
   message <- if (is.na(delim)) {
@@ -1383,5 +1610,179 @@ last_char <- function(x) {
 }
 
 redact_url <- function(x) {
-  sub("^https://[^/]+@", "https://<auth>@", x)
+  sub("://[^/]+@", "://<auth>@", x)
+}
+
+git_dumb_list_refs <- function(url) {
+  synchronize(async_git_dumb_list_refs(url))
+}
+
+async_git_dumb_list_refs <- function(url) {
+  url
+
+  url1 <- paste0(url, "/info/refs")
+  url2 <- paste0(url, "/HEAD")
+  headers <- c(
+    "User-Agent" = git_ua()
+  )
+  when_all(
+    http_get(url1, headers = headers)$
+      then(http_stop_for_status),
+    http_get(url2, headers = headers)$
+      then(http_stop_for_status)
+  )$
+    then(function(res) async_git_dumb_list_refs_process(res, url))
+}
+
+async_git_dumb_list_refs_process <- function(res, url) {
+  res_refs <- res[[1]]
+  res_head <- res[[2]]
+  lines <- strsplit(rawToChar(res_refs$content), "\n", fixed = TRUE)[[1]]
+  lines2 <- strsplit(lines, "\t", fixed = TRUE)
+  ref <- vcapply(lines2, "[[", 2)
+  hash <-   vcapply(lines2, "[[", 1)
+
+  head <- trimws(rawToChar(res_head$content))
+  head <- sub("^[^ ]* ", "", head)
+  has_head <- head != ""
+  if (has_head && ! head %in% ref) {
+    throw(pkg_error(
+      "HEAD does not refer to a ref in git repo at {.url {url}}.",
+      "i" = "HEAD is {.val {head}}."
+    ))
+  }
+
+  list(
+    refs = data_frame(
+      ref = c(if (has_head) "HEAD", ref),
+      hash = c(if (has_head) hash[match(head, ref)], hash)
+    ),
+    caps = character()
+  )
+}
+
+git_dumb_download_file <- function(url, sha, path, output = basename(path)) {
+  invisible(synchronize(async_git_dumb_download_file(url, sha, path, output)))
+}
+
+# This only works for blobs at the root of the tree currently!
+# Also, it only works if the object with sha is not in a pack file!
+# So it is pretty limited, and can only be used with a fallback.
+
+async_git_dumb_download_file <- function(url, sha, path, output = basename(path)) {
+  async_git_dumb_get_commit(url, sha)$
+    then(function(cmt) {
+      async_git_dumb_get_tree(url, cmt[["tree"]])
+    })$
+    then(function(tree) {
+      wh <- match(path, tree$path)
+      if (is.na(wh)) {
+        throw(pkg_error(
+          "Could not find path {.val {path}} in git repo at {.url {url}}."
+        ))
+      }
+      if (tree$type[wh] != "blob") {
+        throw(pkg_error(
+          "Path {.val {path}} is not a blob in git repo at {.url {url}}."
+        ))
+      }
+      tree$hash[wh]
+    })$
+    then(function(blob) {
+      async_git_dumb_get_blob(url, blob)
+    })$
+    then(function(bytes) {
+      if (!is.null(output)) {
+        mkdirp(dirname(output))
+        writeBin(bytes, output)
+      }
+      invisible(bytes)
+    })
+}
+
+async_git_dumb_get_commit <- function(url, sha) {
+  url1 <- paste0(
+    url, "/objects/", substr(sha, 1, 2), "/", substr(sha, 3, nchar(sha))
+  )
+  headers <- c(
+    "User-Agent" = git_ua(),
+    "accept-encoding" = "deflate, gzip"
+  )
+  http_get(url = url1, headers = headers)$
+    then(http_stop_for_status)$
+    then(function(res) {
+      cmt <- zip::inflate(res$content)$output
+      if (any(utils::head(cmt, 6) != charToRaw("commit"))) {
+        throw(pkg_error(
+          "Git object {.val {substr(sha, 1, 7)}} is not a commit object
+           in git repo at {.url {url}}."
+        ))
+      }
+      nul <- which(cmt == 0)
+      if (length(nul) != 1) {
+        throw(pkg_error(
+          "Git commit object {.val {substr(sha, 1, 7)}} is invalid, does not
+           contain a single zero byte, in git repo at {.url {url}}."
+        ))
+      }
+      parse_commit(rawToChar(cmt[(nul+1):length(cmt)]))
+    })
+}
+
+async_git_dumb_get_tree <- function(url, sha) {
+  url1 <- paste0(
+    url, "/objects/", substr(sha, 1, 2), "/", substr(sha, 3, nchar(sha))
+  )
+  headers <- c(
+    "User-Agent" = git_ua(),
+    "accept-encoding" = "deflate, gzip"
+  )
+  http_get(url = url1, headers = headers)$
+    then(http_stop_for_status)$
+    then(function(res) {
+      cmt <- zip::inflate(res$content)$output
+      if (any(utils::head(cmt, 4) != charToRaw("tree"))) {
+        throw(pkg_error(
+          "Git object {.val {substr(sha, 1, 7)}} is not a tree object
+           in git repo at {.url {url}}."
+        ))
+      }
+      nul <- which(cmt == 0)[1]
+      if (is.na(nul)) {
+        throw(pkg_error(
+          "Git tree object {.val {substr(sha, 1, 7)}} is invalid, does not
+           contain any zero bytes, in git repo at {.url {url}}."
+        ))
+      }
+      parse_tree(cmt[(nul+1):length(cmt)])
+    })
+}
+
+async_git_dumb_get_blob <- function(url, sha) {
+  url1 <- paste0(
+    url, "/objects/", substr(sha, 1, 2), "/", substr(sha, 3, nchar(sha))
+  )
+  headers <- c(
+    "User-Agent" = git_ua(),
+    "accept-encoding" = "deflate, gzip"
+  )
+  http_get(url = url1, headers = headers)$
+    then(http_stop_for_status)$
+    then(function(res) {
+      cmt <- zip::inflate(res$content)$output
+      if (any(utils::head(cmt, 4) != charToRaw("blob"))) {
+        throw(pkg_error(
+          "Git object {.val {substr(sha, 1, 7)}} is not a blob object
+           in git repo at {.url {url}}."
+        ))
+      }
+      nul <- which(cmt == 0)[1]
+      if (is.na(nul)) {
+        throw(pkg_error(
+          "Git blob object {.val {substr(sha, 1, 7)}} is invalid, does not
+           contain any zero bytes, in git repo at {.url {url}}."
+        ))
+      }
+      cmt[(nul+1):length(cmt)]
+    })
 }
